@@ -1,3 +1,8 @@
 # -*- coding:utf-8 -*-
 #
 # Copyright (C) 2020, Maximilian Köhl <koehl@cs.uni-saarland.de>
+
+from .checkers import Checker
+
+
+__all__ = ["Checker"]
