@@ -11,7 +11,7 @@ import warnings
 
 from momba import model
 from momba.model import (
-    effects,
+    functions,
     automata,
     context,
     distributions,
@@ -21,9 +21,39 @@ from momba.model import (
     types,
 )
 
+from .dump_model import ModelFeature
 
-class InvalidJANIError(Exception):
-    pass
+
+class JANIError(Exception):
+    """
+    Indicates an error while loading a JANI string.
+
+    Usually either :class:`InvalidJANIError` or
+    :class:`UnsupportedJANIError` gets thrown.
+    """
+
+
+class InvalidJANIError(JANIError):
+    """
+    String is not valid JANI.
+    """
+
+
+class UnsupportedJANIError(JANIError):
+    """
+    Model contains unsupported JANI features.
+
+    Attributes
+    ----------
+    unsupported_features:
+        Used model features not supported by Momba.
+    """
+
+    unsupported_features: t.AbstractSet[t.Union[str, ModelFeature]]
+
+    def __init__(self, unsupported_features: t.AbstractSet[t.Union[str, ModelFeature]]):
+        self.unsupported_features = unsupported_features
+        super().__init__(f"unsupported JANI features {self.unsupported_features!r}")
 
 
 _TYPE_MAP = {
@@ -42,10 +72,10 @@ _BINARY_OP_MAP: t.Mapping[str, expressions.BinaryConstructor] = {
     "⇔": expressions.logic_equiv,  # defined by the `x-momba-operators` extension
     "=": expressions.equals,
     "≠": expressions.not_equals,
-    "<": expressions.less_than,
-    "≤": expressions.less_or_equal_than,
-    ">": expressions.greater_than,  # defined by the `derived-operators` extension
-    "≥": expressions.greater_or_equal_than,  # defined by the `derived-operators` extension
+    "<": expressions.less,
+    "≤": expressions.less_or_equal,
+    ">": expressions.greater,  # defined by the `derived-operators` extension
+    "≥": expressions.greater_or_equal,  # defined by the `derived-operators` extension
     "+": expressions.add,
     "-": expressions.sub,
     "*": expressions.mul,
@@ -54,15 +84,17 @@ _BINARY_OP_MAP: t.Mapping[str, expressions.BinaryConstructor] = {
     "//": expressions.floor_div,  # defined by the `x-momba-operators`extension
     "pow": expressions.power,
     "log": expressions.log,
+    "min": expressions.minimum,  # defined by the `derived-operators` extension
+    "max": expressions.maximum,  # defined by the `derived-operators` extension
 }
 
 _UNARY_OP_MAP: t.Mapping[str, expressions.UnaryConstructor] = {
     "¬": expressions.logic_not,
     "floor": expressions.floor,
     "ceil": expressions.ceil,
-    "abs": expressions.absolute,
-    "trc": expressions.trunc,
-    "sgn": expressions.sgn,
+    "abs": expressions.absolute,  # defined by the `derived-operators` extension
+    "trc": expressions.trunc,  # defined by the `derived-operators` extension
+    "sgn": expressions.sgn,  # defined by the `derived-operators` extension
 }
 
 
@@ -107,9 +139,23 @@ _BINARY_PATH_OPERATORS: t.Mapping[str, operators.BinaryPathOperator] = {
     "R": operators.BinaryPathOperator.RELEASE,
 }
 
+_UNARY_PATH_OPERATORS: t.Mapping[str, operators.UnaryPathOperator] = {
+    "F": operators.UnaryPathOperator.EVENTUALLY,
+    "G": operators.UnaryPathOperator.GLOBALLY,
+}
+
+_TRIGONOMETRIC_FUNCTIONS: t.Mapping[str, operators.TrigonometricFunction] = {
+    function.value: function for function in operators.TrigonometricFunction
+}
+
 
 def _expression(jani_expression: t.Any) -> expressions.Expression:
     if isinstance(jani_expression, (float, bool, int)):
+        if (
+            isinstance(jani_expression, float)
+            and int(jani_expression) == jani_expression
+        ):
+            jani_expression = int(jani_expression)
         return expressions.ensure_expr(jani_expression)
     elif isinstance(jani_expression, str):
         return expressions.name(jani_expression)
@@ -127,6 +173,28 @@ def _expression(jani_expression: t.Any) -> expressions.Expression:
                 _check_fields(jani_expression, required={"op", "exp"})
                 operand = _expression(jani_expression["exp"])
                 return _UNARY_OP_MAP[op](operand)
+            elif op in _TRIGONOMETRIC_FUNCTIONS:
+                _check_fields(jani_expression, required={"op", "exp"})
+                operand = _expression(jani_expression["exp"])
+                return expressions.Trigonometric(_TRIGONOMETRIC_FUNCTIONS[op], operand)
+            elif op == "aa":
+                _check_fields(jani_expression, required={"op", "exp", "index"})
+                return expressions.ArrayAccess(
+                    array=_expression(jani_expression["exp"]),
+                    index=_expression(jani_expression["index"]),
+                )
+            elif op == "av":
+                _check_fields(jani_expression, required={"op", "elements"})
+                return expressions.ArrayValue(
+                    tuple(map(_expression, jani_expression["elements"]))
+                )
+            elif op == "ac":
+                _check_fields(jani_expression, required={"op", "var", "length", "exp"})
+                return expressions.ArrayConstructor(
+                    variable=jani_expression["var"],
+                    length=_expression(jani_expression["length"]),
+                    expression=_expression(jani_expression["exp"]),
+                )
             elif op == "ite":
                 _check_fields(jani_expression, required={"op", "if", "then", "else"})
                 condition = _expression(jani_expression["if"])
@@ -137,14 +205,28 @@ def _expression(jani_expression: t.Any) -> expressions.Expression:
                 _check_fields(jani_expression, required={"op", "var"})
                 variable = jani_expression["var"]
                 return expressions.Derivative(variable)
+            elif op == "call":
+                _check_fields(jani_expression, required={"op", "function", "args"})
+                return functions.CallExpression(
+                    function=jani_expression["function"],
+                    arguments=tuple(
+                        _expression(argument) for argument in jani_expression["args"]
+                    ),
+                )
+            elif op == "nondet":
+                _check_fields(jani_expression, required={"op", "var", "exp"})
+                return expressions.Selection(
+                    variable=jani_expression["var"],
+                    condition=_expression(jani_expression["exp"]),
+                )
         elif "distribution" in jani_expression:
-            _check_fields(jani_expression, required={"op", "args", "distribution"})
-            arguments = list(map(_expression, jani_expression["args"]))
+            _check_fields(jani_expression, required={"args", "distribution"})
+            arguments = tuple(map(_expression, jani_expression["args"]))
             distribution = distributions.DistributionType.by_name(
                 jani_expression["distribution"]
             )
             return expressions.Sample(distribution, arguments)
-    raise InvalidJANIError(f"{jani_expression!r} is not a valid JANI expression")
+    return _property(jani_expression)
 
 
 def _property_interval(jani_property: t.Any) -> properties.Interval:
@@ -192,24 +274,20 @@ def _reward_bound(jani_property: t.Any) -> properties.RewardBound:
     )
 
 
-def _property(jani_property: t.Any) -> properties.Property:
-    try:
-        return _expression(jani_property)
-    except InvalidJANIError:
-        pass
+def _property(jani_property: t.Any) -> expressions.Expression:
     if "op" not in jani_property:
         raise InvalidJANIError(f"{jani_property!r} is not a valid JANI property")
     if jani_property["op"] == "filter":
         _check_fields(jani_property, required={"op", "fun", "values", "states"})
         return properties.Aggregate(
             _AGGREGATION_FUNCTIONS[jani_property["fun"]],
-            _property(jani_property["values"]),
-            _property(jani_property["states"]),
+            _expression(jani_property["values"]),
+            _expression(jani_property["states"]),
         )
     if jani_property["op"] in {"Pmin", "Pmax"}:
         _check_fields(jani_property, required={"op", "exp"})
         return properties.Probability(
-            _MIN_MAX_OPERATORS[jani_property["op"]], _property(jani_property["exp"])
+            _MIN_MAX_OPERATORS[jani_property["op"]], _expression(jani_property["exp"])
         )
     if jani_property["op"] in {"∀", "∃"}:
         _check_fields(jani_property, required={"op", "exp"})
@@ -217,7 +295,7 @@ def _property(jani_property: t.Any) -> properties.Property:
             operators.Quantifier.FORALL
             if jani_property["op"] == "∀"
             else operators.Quantifier.EXISTS,
-            _property(jani_property["exp"]),
+            _expression(jani_property["exp"]),
         )
     if jani_property["op"] in {"Emin", "Emax"}:
         _check_fields(
@@ -232,7 +310,7 @@ def _property(jani_property: t.Any) -> properties.Property:
             },
         )
         accumulate: t.Optional[t.FrozenSet[properties.AccumulationInstant]]
-        reachability: t.Optional[properties.Property]
+        reachability: t.Optional[expressions.Expression]
         step_instant: t.Optional[expressions.Expression]
         time_instant: t.Optional[expressions.Expression]
         reward_instants: t.Optional[t.Sequence[properties.RewardInstant]]
@@ -243,7 +321,7 @@ def _property(jani_property: t.Any) -> properties.Property:
         else:
             accumulate = None
         if "reach" in jani_property:
-            reachability = _property(jani_property["reach"])
+            reachability = _expression(jani_property["reach"])
         else:
             reachability = None
         if "step-instant" in jani_property:
@@ -262,7 +340,7 @@ def _property(jani_property: t.Any) -> properties.Property:
             reward_instants = None
         return properties.ExpectedReward(
             operator=_MIN_MAX_OPERATORS[jani_property["op"]],
-            reward=_property(jani_property["exp"]),
+            reward=_expression(jani_property["exp"]),
             accumulate=accumulate,
             reachability=reachability,
             step_instant=step_instant,
@@ -274,22 +352,23 @@ def _property(jani_property: t.Any) -> properties.Property:
         if "accumulate" in jani_property:
             return properties.SteadyState(
                 _MIN_MAX_OPERATORS[jani_property["op"]],
-                _property(jani_property["exp"]),
+                _expression(jani_property["exp"]),
                 jani_property["accumulate"],
             )
         else:
             return properties.SteadyState(
-                _MIN_MAX_OPERATORS[jani_property["op"]], _property(jani_property["exp"])
+                _MIN_MAX_OPERATORS[jani_property["op"]],
+                _expression(jani_property["exp"]),
             )
+    step_bounds: t.Optional[properties.Interval]
+    time_bounds: t.Optional[properties.Interval]
+    reward_bounds: t.Optional[t.Sequence[properties.RewardBound]]
     if jani_property["op"] in {"U", "W", "R"}:
         _check_fields(
             jani_property,
             required={"op", "left", "right"},
             optional={"step-bounds", "time-bounds", "reward-bounds"},
         )
-        step_bounds: t.Optional[properties.Interval]
-        time_bounds: t.Optional[properties.Interval]
-        reward_bounds: t.Optional[t.Sequence[properties.RewardBound]]
         time_operator = _BINARY_PATH_OPERATORS[jani_property["op"]]
         if "step-bounds" in jani_property:
             step_bounds = _property_interval(jani_property["step-bounds"])
@@ -305,8 +384,33 @@ def _property(jani_property: t.Any) -> properties.Property:
             reward_bounds = None
         return properties.BinaryPathFormula(
             time_operator,
-            _property(jani_property["left"]),
-            _property(jani_property["right"]),
+            _expression(jani_property["left"]),
+            _expression(jani_property["right"]),
+            step_bounds,
+            time_bounds,
+            reward_bounds,
+        )
+    if jani_property["op"] in _UNARY_PATH_OPERATORS:
+        _check_fields(
+            jani_property,
+            required={"op", "exp"},
+            optional={"step-bounds", "time-bounds", "reward-bounds"},
+        )
+        if "step-bounds" in jani_property:
+            step_bounds = _property_interval(jani_property["step-bounds"])
+        else:
+            step_bounds = None
+        if "time-bounds" in jani_property:
+            time_bounds = _property_interval(jani_property["time-bounds"])
+        else:
+            time_bounds = None
+        if "reward-bounds" in jani_property:
+            reward_bounds = list(map(_reward_bound, jani_property["reward-bounds"]))
+        else:
+            reward_bounds = None
+        return properties.UnaryPathFormula(
+            _UNARY_PATH_OPERATORS[jani_property["op"]],
+            _expression(jani_property["exp"]),
             step_bounds,
             time_bounds,
             reward_bounds,
@@ -331,6 +435,9 @@ def _type(typ: t.Any) -> types.Type:
         lower_bound = _expression(typ["lower-bound"]) if "lower-bound" in typ else None
         upper_bound = _expression(typ["upper-bound"]) if "upper-bound" in typ else None
         return base.bound(lower_bound, upper_bound)
+    elif typ["kind"] == "array":
+        _check_fields(typ, required={"kind", "base"})
+        return types.array_of(_type(typ["base"]))
     raise InvalidJANIError(f"{typ!r} is not a valid JANI type")
 
 
@@ -358,7 +465,7 @@ def _check_fields(
     jani_structure: t.Any,
     required: _Fields = frozenset(),
     optional: _Fields = frozenset(),
-    unsupported: _Fields = frozenset({"comment"}),
+    unsupported: _Fields = frozenset(),
 ) -> None:
     if not isinstance(jani_structure, dict):
         raise InvalidJANIError(f"expected map but found {jani_structure!r}")
@@ -429,13 +536,13 @@ def _location(jani_location: t.Any) -> automata.Location:
         progress_invariant = _expression(jani_location["time-progress"]["exp"])
     else:
         progress_invariant = None
-    transient_values: t.Set[effects.Assignment] = set()
+    transient_values: t.Set[automata.Assignment] = set()
     if "transient-values" in jani_location:
         for jani_transient_value in jani_location["transient-values"]:
             _check_fields(
                 jani_transient_value, required={"ref", "value"}, optional={"comment"}
             )
-            assignment = effects.Assignment(
+            assignment = automata.Assignment(
                 target=_target(jani_transient_value["ref"]),
                 value=_expression(jani_transient_value["value"]),
             )
@@ -452,9 +559,16 @@ def _location(jani_location: t.Any) -> automata.Location:
 _Locations = t.Dict[str, automata.Location]
 
 
-def _target(jani_target: t.Any) -> effects.Target:
+def _target(jani_target: t.Any) -> expressions.Expression:
     if isinstance(jani_target, str):
-        return effects.Name(jani_target)
+        return expressions.Name(jani_target)
+    _check_fields(jani_target, required={"op"}, optional={"exp", "index"})
+    if jani_target["op"] == "aa":
+        _check_fields(jani_target, required={"op", "exp", "index"})
+        return expressions.ArrayAccess(
+            array=_expression(jani_target["exp"]),
+            index=_expression(jani_target["index"]),
+        )
     raise InvalidJANIError(f"{jani_target!r} is not a valid lvalue")
 
 
@@ -467,7 +581,7 @@ def _edge(ctx: model.Context, locations: _Locations, jani_edge: t.Any) -> automa
     location = locations[jani_edge["location"]]
     action_pattern: t.Optional[model.ActionPattern] = None
     if "action" in jani_edge:
-        if isinstance(str, jani_edge["action"]):
+        if isinstance(jani_edge["action"], str):
             action_pattern = model.ActionPattern(
                 ctx.get_action_type_by_name(jani_edge["action"])
             )
@@ -491,12 +605,12 @@ def _edge(ctx: model.Context, locations: _Locations, jani_edge: t.Any) -> automa
                 else None
             ),
             assignments=frozenset(
-                effects.Assignment(
+                automata.Assignment(
                     target=_target(jani_assignment["ref"]),
                     value=_expression(jani_assignment["value"]),
                     index=jani_assignment.get("index", 0),
                 )
-                for jani_assignment in jani_destination["assignments"]
+                for jani_assignment in jani_destination.get("assignments", ())
             ),
         )
         for jani_destination in jani_edge["destinations"]
@@ -535,19 +649,61 @@ def _action(jani_action: t.Any) -> model.ActionType:
 JANIModel = t.Union[bytes, str]
 
 
-def load_model(source: JANIModel) -> model.Network:
-    """
-    Constructs a Momba automata network based on the provided JANI model.
+JSONObject = t.Mapping[str, t.Any]
 
-    :param source:
-        The source of the JANI model to load.
-    :return:
-        The resulting network of Momba automata network.
+
+def _load_parameter(jani_parameter: JSONObject) -> functions.FunctionParameter:
+    _check_fields(jani_parameter, required={"name", "type"})
+    return functions.FunctionParameter(
+        name=jani_parameter["name"], typ=_type(jani_parameter["type"])
+    )
+
+
+def _load_functions(
+    scope: context.Scope, jani_functions: t.Sequence[JSONObject]
+) -> None:
+    for jani_function in jani_functions:
+        _check_fields(jani_function, required={"name", "type", "parameters", "body"})
+        scope.define_function(
+            name=jani_function["name"],
+            parameters=[
+                _load_parameter(jani_parameter)
+                for jani_parameter in jani_function["parameters"]
+            ],
+            returns=_type(jani_function["type"]),
+            body=_expression(jani_function["body"]),
+        )
+
+
+def load_model(source: JANIModel, *, ignore_properties: bool = False) -> model.Network:
+    """
+    Constructs a Momba automaton :class:`~momba.model.Network` based on
+    the provided JANI model string.
+
+    The flag `ignore_properties` indicates whether the properties of the
+    model should be ignored or added to the returned network's modeling
+    :class:`~momba.model.Context`.
+
+    Throws a :class:`~momba.jani.JANIError` in case the JANI file cannot be
+    loaded.
     """
     if isinstance(source, bytes):
         jani_model = json.loads(source.decode("utf-8"))
     else:
         jani_model = json.loads(source)
+    features = {ModelFeature(feature) for feature in jani_model.get("features", [])}
+    unsupported_features = features.difference(
+        {
+            ModelFeature.ARRAYS,
+            ModelFeature.DERIVED_OPERATORS,
+            ModelFeature.FUNCTIONS,
+            ModelFeature.STATE_EXIT_REWARDS,
+            ModelFeature.NONDET_SELECTION,
+            ModelFeature.TRIGONOMETRIC_FUNCTIONS,
+        }
+    )
+    if unsupported_features:
+        raise UnsupportedJANIError(unsupported_features)
     _check_fields(
         jani_model,
         required={"jani-version", "name", "type", "automata", "system"},
@@ -560,18 +716,22 @@ def load_model(source: JANIModel) -> model.Network:
             "restrict-initial",
             "properties",
             "comment",
+            "functions",
         },
         unsupported={"comment"},
     )
-    network = model.Network(model.Context())
+
+    network = model.Network(model.Context(model.ModelType[jani_model["type"].upper()]))
     if "variables" in jani_model:
         for jani_declaration in jani_model["variables"]:
             var_declaration = _variable_declaration(jani_declaration)
-            network.ctx.global_scope.add_declaration(var_declaration)
+            network.ctx.global_scope.add_declaration(var_declaration, validate=False)
     if "constants" in jani_model:
         for jani_declaration in jani_model["constants"]:
             const_declaration = _constant_declaration(jani_declaration)
-            network.ctx.global_scope.add_declaration(const_declaration)
+            network.ctx.global_scope.add_declaration(const_declaration, validate=False)
+    for declaration in network.ctx.global_scope.declarations:
+        declaration.validate(network.ctx.global_scope)
     if "restrict-initial" in jani_model:
         _check_fields(
             jani_model["restrict-initial"], required={"exp"}, optional={"comment"}
@@ -580,12 +740,19 @@ def load_model(source: JANIModel) -> model.Network:
         network.initial_restriction = initial_restriction
     if "actions" in jani_model:
         for jani_action in jani_model["actions"]:
-            network.ctx.add_action_type(_action(jani_action))
+            network.ctx._add_action_type(_action(jani_action))
+    _load_functions(network.ctx.global_scope, jani_model.get("functions", ()))
     for jani_automaton in jani_model["automata"]:
         _check_fields(
             jani_automaton,
             required={"name", "locations", "initial-locations", "edges"},
-            optional={"variables", "restrict-initial", "comment", "x-momba-anonymous"},
+            optional={
+                "variables",
+                "functions",
+                "restrict-initial",
+                "comment",
+                "x-momba-anonymous",
+            },
         )
         name: t.Optional[str]
         if jani_automaton.get("x-momba-anonymous", False):
@@ -596,11 +763,14 @@ def load_model(source: JANIModel) -> model.Network:
         if "variables" in jani_automaton:
             for jani_declaration in jani_automaton["variables"]:
                 declaration = _variable_declaration(jani_declaration)
-                automaton.scope.add_declaration(declaration)
+                automaton.scope.add_declaration(declaration, validate=False)
+        for declaration in automaton.scope.declarations:
+            declaration.validate(automaton.scope)
         locations = {
             jani_location["name"]: _location(jani_location)
             for jani_location in jani_automaton["locations"]
         }
+        _load_functions(automaton.scope, jani_automaton.get("functions", ()))
         if "restrict-initial" in jani_automaton:
             _check_fields(
                 jani_automaton["restrict-initial"],
@@ -609,16 +779,50 @@ def load_model(source: JANIModel) -> model.Network:
             )
             initial_restriction = _expression(jani_automaton["restrict-initial"]["exp"])
             automaton.initial_restriction = initial_restriction
+        for location in locations.values():
+            automaton.add_location(location)
         for jani_edge in jani_automaton["edges"]:
             automaton.add_edge(_edge(network.ctx, locations, jani_edge))
         for jani_location in jani_automaton["initial-locations"]:
-            automaton.add_initial_location(locations[jani_location])
-    for jani_prop in jani_model["properties"]:
+            automaton.add_location(locations[jani_location], initial=True)
+    jani_system = jani_model["system"]
+    instances: t.List[model.Instance] = []
+    _check_fields(jani_system, required={"elements"}, optional={"syncs", "comment"})
+    for element in jani_system["elements"]:
         _check_fields(
-            jani_prop,
-            required={"expression", "name"},
+            element, required={"automaton"}, optional={"input-enable", "comment"}
         )
-        network.ctx.define_property(
-            jani_prop["name"], _property(jani_prop["expression"])
+        automaton = network.ctx.get_automaton_by_name(element["automaton"])
+        input_enable = element.get("input-enable", None)
+        instances.append(
+            automaton.create_instance(
+                input_enable=frozenset()
+                if input_enable is None
+                else frozenset(input_enable)
+            )
         )
+    if "syncs" in jani_system:
+        for jani_vector in jani_system["syncs"]:
+            vector: t.Dict[model.Instance, model.ActionPattern] = {}
+            for index, action_name in enumerate(jani_vector["synchronise"]):
+                if action_name is None:
+                    continue
+                instance = instances[index]
+                action_type = network.ctx.get_action_type_by_name(action_name)
+                vector[instance] = action_type.create_pattern()
+            result_pattern = None
+            if "result" in jani_vector:
+                result_pattern = network.ctx.get_action_type_by_name(
+                    jani_vector["result"]
+                ).create_pattern()
+            network.create_link(vector, result=result_pattern)
+    if not ignore_properties:
+        for jani_prop in jani_model["properties"]:
+            _check_fields(
+                jani_prop,
+                required={"expression", "name"},
+            )
+            network.ctx.define_property(
+                jani_prop["name"], _property(jani_prop["expression"])
+            )
     return network
